import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.util.Version;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.io.IOException;


public class QASearch {
    public static void readQueriesAndSearch(String fileName, IndexSearcher searcher, int maxHits){
        BufferedReader buf = null;
        try {
            buf = new BufferedReader(new FileReader(fileName));
            String line = null;                    
            Version v = Version.LUCENE_35;
            String defaultField = "text";
            // we use the same StandardAnalyzer we used for indexing
            Analyzer analyzer = new StandardAnalyzer(v);
            QueryParser parser = new QueryParser(v, defaultField, analyzer);

            int count = 0;
            while((line = buf.readLine()) != null) {
                ++count;
                // split the input into query id and query itself
                String[] tokens = line.trim().split("\t");
                String qid = tokens[0];
                String qtext = tokens[1];   
                
                // Lucene uses special characters: + - && || ! ( ) { } [ ] ^ " ~ * ? : \ 
                // as part of its query language, so we need to escape them
                // one quick and dirty solution is to simply remove all non-alphanumeric characters from the query
                qtext = qtext.replaceAll("[^A-Za-z0-9]", " ");
                // System.err.printf("%s %s\n", qid, qtext);                
                Query q = parser.parse(qtext);

                // search for a query retrieving maxHits results
                TopDocs hits = searcher.search(q, maxHits);
                ScoreDoc[] scoreDocs = hits.scoreDocs;
                
                // write the results to standard output in the following format:
                // <question id>   <answer id>  <score of the SE>
                for (int n = 0; n < scoreDocs.length; ++n) {
                    ScoreDoc sd = scoreDocs[n];
                    float score = sd.score;
                    int docId = sd.doc;
                    Document d = searcher.doc(docId);
                    String aid = d.get("id");

                    System.out.printf("%s\t%s\t%.8f\n", qid, aid, score);                    
                    // System.out.printf("%3d %.4f  %s  %s\n", n, score, aid, atext);
                }                
                if (count % 100 == 0){
                    System.err.printf("%d..", count);
                }
            }  
            System.err.printf("done!\n");
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ParseException e) {
            e.printStackTrace();
        } finally {
            try {
                if (buf != null)
                    buf.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }       
    }
    
    public static void main(String[] args) 
        throws CorruptIndexException,
               IOException {

        // path to the directory containing Lucene index
        File indexDir = new File(args[0]);
        
        // file containing the queries
        String queryFileName = args[1];

        // maximum number of documents to retrieve for each query
        int maxHits = Integer.parseInt(args[2]);

        System.err.println("Index Dir = " + indexDir.getCanonicalPath());
        
        // open the index
        Directory dir = FSDirectory.open(indexDir);
        IndexReader reader = IndexReader.open(dir);

        // instantiate a searcher
        IndexSearcher searcher = new IndexSearcher(reader);

        // process queries and search
        readQueriesAndSearch(queryFileName, searcher, maxHits);        
    }

}