#! /usr/bin/env python
from __future__ import division
import sys
from collections import defaultdict
from operator import itemgetter
from optparse import OptionParser


class ResFileReader(object):
	"""
	Use this class to handle various file formats coming from 
	diffeernt IR engines, i.e. Terrier (answerbag) and Watson (jeopardy)
	The information we need to extract is qid, if the candidate answer is relevant or not 
	and the score IR engine assigned to this answer

	FIX: pretty up this class later on.	
	"""
	def __init__(self, mode):
		self.process_line = getattr(self, "process_line_%s" %mode)

	def process_line_jeopardy(self, line):
		tokens = line.strip().split("\t")
		qid = tokens[0]
		aid = tokens[1]
		ir_score = 1.0/(int(tokens[2]) + 1)  # we invert the score
		relevant = tokens[3]  # true or false
		return qid, aid, relevant, ir_score

	def process_line_answerbag(self, line):
		tokens = line.strip().split("\t")
		qid = tokens[0]
		aid = tokens[1]
		relevant = "false"
		if qid == aid:
			relevant = "true"		
		ir_score = float(tokens[2])	
		return qid, aid, relevant, ir_score


def read_res_file(res_fname, mode="jeopardy"):
	res = open(res_fname).readlines()

	process_line = ResFileReader(mode).process_line

	ir = defaultdict(list)
	for line_res in res:
		qid, aid, relevant, ir_score = process_line(line_res)  # process the line from the res file
		ir[qid].append( (relevant, ir_score) )
	
	for qid in ir:
		ir[qid] = [rel for rel, score in sorted(ir[qid], key = itemgetter(1), reverse = True)]  # sort based on the score
	return ir


def read_res_pred_files(res_fname, pred_fname, mode="jeopardy", verbose=True):
	res = open(res_fname).readlines()
	pred = open(pred_fname).readlines()

	process_line = ResFileReader(mode).process_line

	ir, svm = defaultdict(list), defaultdict(list)
	for line_res, line_pred in zip(res, pred):
		qid, aid, relevant, ir_score = process_line(line_res)  # process the line from the res file
		pred_score = float(line_pred.strip())
		ir[qid].append( (relevant, ir_score, aid) )
		svm[qid].append( (relevant, pred_score, aid) )
	
	if verbose:
		rer_fname = res_fname + ".reranking.txt"
		file_reranking = open(rer_fname, "w")
		print "Reranking output is written to: %s" % rer_fname
		# print "%5s %8s  %2s %2s" % ("qid", "aid", "ir", "svm")
		# print "="*21

	for qid in ir:
		ir_sorted = sorted(ir[qid], key = itemgetter(1), reverse = True)
		svm_sorted = sorted(svm[qid], key = itemgetter(1), reverse = True)
		if verbose:
			ir_sorted_by_aid  = sorted([(aid, rank) for rank, (rel, score, aid) in enumerate(ir_sorted) if rel == "true"], key=itemgetter(0))
			svm_sorted_by_aid = sorted([(aid, rank) for rank, (rel, score, aid) in enumerate(svm_sorted) if rel == "true"], key=itemgetter(0))
			for (ir_aid, ir_rank), (svm_aid, svm_rank) in zip(ir_sorted_by_aid, svm_sorted_by_aid):
				file_reranking.write("%s %s  %2s %2s\n" % (qid, ir_aid, ir_rank, svm_rank))
			# print "-"*21

		ir[qid] = [rel for rel, score, aid in sorted(ir[qid], key=itemgetter(1), reverse=True)]  # sort based on the score
		svm[qid] = [rel for rel, score, aid in sorted(svm[qid], key=itemgetter(1), reverse=True)]  # sort based on the score
	return ir, svm


def compute_mrr(out):
	n = len(out)
	MRR = 0.0
	for qid in out:
		relevant = out[qid]
		for i, rel in enumerate(relevant, 1):
			if rel == "true":
				MRR += 1.0/i
	return MRR*100/n


def compute_map(out, th):
	n = len(out)
	MAP = 0.0
	for qid in out:
		relevant = out[qid]
		# compute the number of relevant docs
		# get a list of precisions in the range(0,th)
		prec = [0.0]*th
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				prec[i] += 1.0
		for i in xrange(1, th):
			prec[i] += prec[i-1]	
		# print num_relevant, sum([p/i for i, p in enumerate(prec, 1)])/num_relevant, [p/i for i, p in enumerate(prec, 1)]
		# print
		MAP += sum([p/i for i, p in enumerate(prec, 1)])/th

	return MAP/n


def compute_precision(out, th):
	precisions = [0.0]*th
	for qid in out:
		relevant = out[qid]
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				precisions[i] += 1.0
				break
	for i in xrange(1, th):
		precisions[i] += precisions[i-1]	

	return [p*100/len(out) for p in precisions]

	
def compute_accuracy(out, th):
	acc = [0.0]*th
	for qid in out:
		relevant = out[qid]
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				acc[i] += 1.0
	for i in xrange(1, th):
		acc[i] += acc[i-1]	
	
	return [a*100/(i*len(out)) for i, a in enumerate(acc, 1)]


def eval_reranker(res_fname="svm.test.res", pred_fname="svm.train.pred", th=50, mode="jeopardy", verbose=False):
	try:
		ir, svm = read_res_pred_files(res_fname, pred_fname, mode, verbose)		
	except:
		print "Failed to process input files: %s %s" % (res_fname, pred_fname)
		print "Check that the input file format is correct"
		sys.exit(1)

	if mode == "jeopardy":
			# evaluate IR
		prec_se = compute_precision(ir, th)
		acc_se = compute_accuracy(ir, th)

		# evaluate SVM
		prec_svm = compute_precision(svm, th)
		acc_svm = compute_accuracy(svm, th)

		map_se = compute_map(ir, th)
		map_svm = compute_map(svm, th)
		print "MAP: %5.2f %5.2f" %(map_se, map_svm)
		print
		print "%15s  %5s  %13s  %5s" % ("IR", "SVM", "IR", "SVM")
		for i, (p_se, p_svm, a_se, a_svm) in enumerate(zip(prec_se, prec_svm, acc_se, acc_svm), 1):
			print "REC-1@%02d: %5.2f  %5.2f  ACC@%02d: %5.2f  %5.2f" %(i, p_se, p_svm, i, a_se, a_svm)

	else: # answerbag
		prec_se = compute_precision(ir, th)
		prec_svm = compute_precision(svm, th)
		mrr_svm = compute_mrr(svm)
		mrr_se = compute_mrr(ir)
		print "%15s  %5s" % ("IR", "SVM")
		print "%-8s: %5.2f  %5.2f" %("MRR", mrr_se, mrr_svm)
		for i, (p_se, p_svm) in enumerate(zip(prec_se, prec_svm), 1):
			print "REC-1@%02d: %5.2f  %5.2f" % (i, p_se, p_svm)
	


def eval_search_engine(res_fname="svm.test.res", th=50, mode="jeopardy"):
	try:	
		ir = read_res_file(res_fname, mode)		
	except:
		print "Failed to process input file: %s" % res_fname
		print "Check that the input file format is correct"
		sys.exit(1)

	# evaluate IR
	prec = compute_precision(ir, th)
	acc = compute_accuracy(ir, th)

	if mode == "jeopardy":	# for jeopardy data we compute MAP instead of MRR since here for each query we have multiple correct answers
		MAP = compute_map(ir, th)
		print "MAP: %5.2f" % MAP
	else:  # answerbag
		MRR = compute_mrr(ir)
		print "MRR^: %5.2f" % MRR
	
	# print stats
	print
	print "%2s: %5s  %5s" % ("#", "REC-1", "ACC")
	for i, (p, a) in enumerate(zip(prec, acc), 1):
		print "%02d: %5.2f  %5.2f" % (i, p, a)



def main():
	usage = "usage: %prog [options] arg1 [arg2]"
	desc = """arg1: file with the output of the baseline search engine (ex: svm.test.res) 
	arg2: predictions file from svm (ex: train.predictions)
	if arg2 is ommitted only the accuracy of the search engine is evaluated"""

	parser = OptionParser(usage=usage, description=desc)
	parser.add_option("-t", "--threshold", dest="th", default=10, type=int,
	                  help="supply a value for computing Precision up to a given threshold [default: %default]", metavar="VALUE")	 	  
	parser.add_option("-m", "--mode", dest="mode", default="answerbag", 
                  help="evalute jeopardy or answerbag [default: %default]")	 	  
	parser.add_option("-v", "--verbose", dest="verbose", default=False, action="store_true",
                  help="produce verbose output [default: %default]")	 	  
		                  

	(options, args) = parser.parse_args()

	if len(args) == 1:
		res_fname = args[0]
		eval_search_engine(res_fname, options.th, options.mode)
	elif len(args) == 2:
		res_fname = args[0]
		pred_fname = args[1]	
		eval_reranker(res_fname, pred_fname, options.th, options.mode, options.verbose)
	else:
		parser.print_help()
		sys.exit(1)
	

if __name__ == '__main__':	
	main()
